
import datetime
import pytest
import secrets
import tempfile


from starfish.agent.memory_agent import MemoryAgent
from starfish.asset.data_asset import DataAsset
from starfish.account import Account

VALID_DID = 'did:dep:' + secrets.token_hex(64)
INVALID_DID = 'did:dep:' + secrets.token_hex(128)


def create_agent(network):
    agent = MemoryAgent(network)
    assert(agent)
    return agent

def register_asset(agent):
    asset = DataAsset.create('test memory agent asset', secrets.token_bytes(1024))
    assert(asset)
    asset = agent.register_asset(asset)
    assert(asset)
    return asset

def create_listing(agent, resources, asset):
    listing = agent.create_listing(resources.listing_data, asset.did)
    assert(listing)
    return (listing)

def purchase_asset(network, resources, config):
    agent = create_agent(network)
    asset = register_asset(agent)
    listing = create_listing(agent, resources, asset)

    account = Account(config.accounts[1].as_dict)
    purchase = agent.purchase_asset(listing, account)
    return purchase, listing, agent, asset, account

def test_init(network):
    agent = MemoryAgent(network)
    assert(agent)

def test_register_asset(network, resources, config):
    agent = create_agent(network)
    asset = register_asset(agent)
    listing = create_listing(agent, resources, asset)
    assert(listing)
    assert(listing.listing_id)


def test_get_listing(network, resources, config):
    agent = create_agent(network)
    asset = register_asset(agent)
    listing = create_listing(agent, resources, asset)
    found_listing = agent.get_listing(listing.listing_id)
    assert(found_listing)
    assert(found_listing.listing_id == listing.listing_id)

def test_search_listings(network, resources, config):
    agent = create_agent(network)
    asset = register_asset(agent)
    listing = create_listing(agent, resources, asset)
    listing_ids = agent.search_listings(resources.listing_data['author'])
    assert(listing_ids)
    assert(len(listing_ids) > 0)
    is_found = False
    for listing_id in listing_ids:
        if listing_id == listing.listing_id:
            is_found = True
            break
    assert(is_found)

def test_purchase_asset(network, resources, config):
    agent = create_agent(network)
    asset = register_asset(agent)
    listing = create_listing(agent, resources, asset)
    account = Account(config.accounts[1].as_dict)
    purchase = agent.purchase_asset(listing, account)
    assert(purchase)


def test_is_access_granted_for_asset(network, resources, config):
    purchase, listing, agent, asset, account = purchase_asset(network, resources, config)
    assert(agent.is_access_granted_for_asset(asset, account, purchase.purchase_id,))

def test_consume_asset(network, resources, config):
    purchase, listing, agent, asset, account = purchase_asset(network, resources, config)
    assert(agent.consume_asset(listing, account, purchase.purchase_id,))

def test_is_did_valid():
    assert(MemoryAgent.is_did_valid(VALID_DID))
    # FIXME: This should be invalid
    assert( MemoryAgent.is_did_valid(INVALID_DID))



"""
    Service Class
    To handle service items in a DDO record
"""

import json
import re


class Service:
    """
    Service class to create validate service in a DDO
    """

    def __init__(self, service_id, endpoint, service_type, values):
        self._id = service_id
        self._endpoint = endpoint
        self._type = service_type

        # assign the _values property to empty until they are used
        self._values = {}
        reserved_names = {'id', 'serviceEndpoint', 'type'}
        if values:
            for name, value in values.items():
                if name not in reserved_names:
                    self._values[name] = value

    @property
    def id(self):
        return self._id

    def assign_did(self, did):
        """ Assign a new DID/Id to the service"""
        if re.match('^#.*', self._id):
            self._id = did + self._id

    @property
    def type(self):
        """get the service type"""
        return self._type

    @property
    def endpoint(self):
        """get the service endpoint"""
        return self._endpoint

    @property
    def values(self):
        """get any service value s"""
        return self._values

    def update_value(self, name, value):
        if name not in {'id', 'serviceEndpoint', 'type'}:
            self._values[name] = value

    def as_text(self, is_pretty=False):
        """return the service as a JSON string"""
        values = {
            'id': self._id,
            'type': self._type,
            'serviceEndpoint': self._endpoint
        }
        if self._values:
            # add extra service values to the dictonairy
            for name, value in self._values.items():
                values[name] = value

        if is_pretty:
            return json.dumps(values, indent=4, separators=(',', ': '))

        return json.dumps(values)

    @property
    def as_dictionary(self):
        """return the service as a python dictionary"""
        values = {
            'id': self._id,
            'type': self._type,
            'serviceEndpoint': self._endpoint
        }
        if self._values:
            # add extra service values to the dictonairy
            for name, value in self._values.items():
                if isinstance(value, object) and hasattr(value, 'as_dictionary'):
                    value = value.as_dictionary()
                elif isinstance(value, list):
                    value = [v.as_dictionary() if hasattr(v, 'as_dictionary') else v for v in value]

                values[name] = value
        return values

    @property
    def is_valid(self):
        """return True if the sevice is valid"""
        return self._endpoint is not None and self._type is not None



"""

Unit test AgentManager

"""
import pytest
import secrets

from starfish.agent import AgentManager
from starfish.asset import DataAsset


def test_agent_manager_add(config, network):
    manager = AgentManager(network)

    for name, item in config.agent_list.items():
        manager.add(name, item)

    assert(manager.items)
    for name, item in manager.items.items():
        assert(config.agent_list[name])


    for name, item in config.agent_list.items():
        ddo = manager.resolve_ddo(name)
        assert(ddo)

def test_agent_manager_load_agent(config, network):
    manager = AgentManager(network)

    for name, item in config.agent_list.items():
        manager.add(name, item)

    items = config.agent_list;
    name = list(items.keys())[0]
    ddo_text = manager.resolve_ddo(name)
    ddo = AgentManager.create_ddo_object(ddo_text)

    # load a named item
    remote_agent = manager.load_agent(name)
    assert(remote_agent)
    assert(remote_agent.did == ddo.did)

    # load from a agent did
    remote_agent = manager.load_agent(ddo.did)
    assert(remote_agent)
    assert(remote_agent.did == ddo.did)


    test_data = secrets.token_hex(1024)
    asset_data = DataAsset.create('TestAsset', test_data)
    asset = remote_agent.register_asset(asset_data)
    assert(asset)

    # load from a asset_did
    remote_agent = manager.load_agent(asset.did)
    assert(remote_agent)
    assert(remote_agent.did == ddo.did)




"""

    Account class to provide basic functionality for all Starfish library accounts.

"""

from web3 import Web3


class Account():
    """

    Account class, adds functionality for an account to be used on the starfish named network.

    :param address: address or dict of the account details
    :type address: string or dict
    :param password: password for the account
    :type password: str or None
    :param keyfile: keyfile containing the private encrypted key
    :type keyfile: str or None

    If the address parameter is a string then it's the account address.
    If dict then the dict can be the following format: ::

        {
            'address': 'xxxx',
            'password': 'yyyy',
        }

    """

    def __init__(self, address, password=None, keyfile=None):
        """init a standard ocean agent"""
        self._address = None
        self._password = None

        if isinstance(address, dict):
            self._address = Web3.toChecksumAddress(address.get('address'))
            self._password = address.get('password')
            self._keyfile = address.get('keyfile')
        elif isinstance(address, (tuple, list)):
            self._address = Web3.toChecksumAddress(address[0])
            if len(address) > 1:
                self._password = address[1]
            if len(address) > 2:
                self._keyfile = address[2]
        elif isinstance(address, str):
            self._address = Web3.toChecksumAddress(address)
            self._password = password
            self._keyfile = keyfile

    def is_address_equal(self, address):
        """

        Compares two addresses if are equal. Both addresses are converted to checksum
        address and then compared.

        :param str address: address to compare with this object's address
        :return: True if the param address is the same is the one held in this account
        :type: boolean

        >>> account = ocean.get_account('0x00bd138abd70e2f00903268f3db08f2d25677c9e')
        >>> account.is_address_equal('0x00Bd138aBD70e2F00903268F3Db08f2D25677C9e')
        True
        """
        return self.as_checksum_address == Web3.toChecksumAddress(address)

    def sign_transaction(self, web3, transaction):
        encrypted_key = None
        with open(self._keyfile, 'r') as fp:
            encrypted_key = fp.read()
        if encrypted_key:
            secret_key = web3.eth.account.decrypt(encrypted_key, self._password)
            signed = web3.eth.account.sign_transaction(transaction, secret_key)
            return signed

    @property
    def is_password(self):
        """
        Return True if the password has been set, else return False

        """
        return self._password is not None

    @property
    def address(self):
        """

        Return the account address for this account

        :return: address
        :type: str

        >>> account = ocean.get_account('0x00bd138abd70e2f00903268f3db08f2d25677c9e')
        >>> account.address
        0x00bd138abd70e2f00903268f3db08f2d25677c9e
        """
        return self._address

    @property
    def as_checksum_address(self):
        """

        Return the address as a checksum address

        :return: checksum address
        :type: str

        >>> account = ocean.get_account('0x00bd138abd70e2f00903268f3db08f2d25677c9e')
        >>> account.as_checksum_address
        0x00Bd138aBD70e2F00903268F3Db08f2D25677C9e

        """
        if self._address:
            return Web3.toChecksumAddress(self._address.lower())
        return None

    @property
    def password(self):
        """

        Return the account password for this account

        :return: password
        :type: str

        >>> account.password
        secret

        """
        return self._password

    def set_password(self, password):
        """

        Set the password for this account

        :param str password: Password to set for this account

        >>> account.set_password('new secret')
        >>> account.password
        new secret
        """
        self._password = password

    @property
    def keyfile(self):
        return self._keyfile

    @property
    def is_valid(self):
        return self._address and self._password and self._keyfile

    def __str__(self):
        return f'Account: {self.address}'




import pytest


from starfish.agent.services import (
    Services,
    SERVICES,
    ALL_SERVICES
)


def test_add_single_default():
    # I want to add a meta service using the default url & version
    services = Services('http://localhost')
    services.add('meta')
    result = services.as_dict
    assert(result)
    assert('meta' in result)
    assert(result['meta']['url'] == 'http://localhost/api/v1/meta')


def test_add_single_version_default():
    # I want to add meta service using the a default version
    services = Services('http://localhost', version='v99')
    services.add('meta')
    result = services.as_dict
    assert(result)
    assert('meta' in result)
    assert(result['meta']['url'] == 'http://localhost/api/v99/meta')
    assert('meta' in services.names)


def test_add_multiple():
    # I want to add only meta and storage services
    services = Services('http://localhost', service_list=['meta', 'storage'])
    result = services.as_dict
    assert(result)
    assert('meta' in result)
    assert(result['meta']['url'] == 'http://localhost/api/v1/meta')
    assert('storage' in result)
    assert(result['storage']['url'] == 'http://localhost/api/v1/assets')


def test_add_all():
    # I want to add all default services
    services = Services('http://localhost', all_services=True)
    result = services.as_dict
    assert(result)
    assert(services.names == ALL_SERVICES)

def test_add_unknown_default():
    # I want to see when a add an unkown service name using the default serivce type
    services = Services('http://localhost', all_services=True)
    result = services.as_dict
    with pytest.raises(ValueError):
        services.add('bad-name')

def test_add_new_service():
    # I want to add a new service uri and service type
    services = Services('http://localhost', all_services=True)
    services.add('test', 'api/v1/test', 'NEW.v1.test')
    result = services.as_dict
    assert(result)
    assert(result['test']['type'] == 'NEW.v1.test')
    assert(result['test']['url'] == 'http://localhost/api/v1/test')



"""

    Direct Purchase Contract

"""

from .contract_base import ContractBase

CONTRACT_NAME = 'Dispenser'


class DispenserContract(ContractBase):
    """Class representing the Ocean token dispenser contract."""

    def __init__(self):
        ContractBase.__init__(self, CONTRACT_NAME)

    def request_tokens(self, account, amount):
        # amount is in ether
        return self.call('requestTokens', amount, account)



"""


Test bundle file upload/download


"""

import hashlib
import tempfile
import secrets

from starfish.utils.data_bundle import (
    decode_readable_size,
    register_upload_bundle_file,
    download_bundle_file
)

TEST_FILE_SIZE = '120mb'

def test_bundle_file_upload_download(remote_agent):
    # create temp large file
    max_size = decode_readable_size(TEST_FILE_SIZE)
    size = 0
    with tempfile.NamedTemporaryFile() as fp:
        md5sum = hashlib.md5()
        while size < max_size:
            data = secrets.token_bytes(1024 * 100)
            size += len(data)
            fp.write(data)
            md5sum.update(data)

        test_data_md5 = md5sum.hexdigest()
        fp.seek(0)
        filename = fp.name
        asset_bundle = register_upload_bundle_file(remote_agent, filename)

        fp.seek(0)

        with tempfile.NamedTemporaryFile() as out_fp:
            download_size = download_bundle_file(remote_agent, asset_bundle, out_fp.name)
            out_fp.seek(0)
            assert(download_size == size)
            md5sum = hashlib.md5()
            while True:
                data = out_fp.read(1024 * 100)
                if data:
                    md5sum.update(data)
                else:
                    break

            assert(test_data_md5 == md5sum.hexdigest())



"""


    Test Contract Manager


"""

from starfish.contract import ContractManager

def test_contract_load(network):
    manager = ContractManager(network.web3, 'starfish.contract')
    contract = manager.load('DirectPurchaseContract', network.name)



"""
    test_13_asset_listing_search

    As a developer building client code to interact with an Ocean marketplace,
    I need a way to search available asset listings

"""

import secrets

from starfish.asset import DataAsset


def test_13_asset_listing_search(resources, remote_agent):
    test_data = secrets.token_hex(1024)
    asset_data = DataAsset.create('TestAsset', test_data)
    asset = remote_agent.register_asset(asset_data)
    assert(asset)
    listing = remote_agent.create_listing(resources.listing_data, asset.did)
    assert(listing)
    assert(listing.asset_did)



"""

    DIDRegistry Contract

"""

import logging


from starfish.contract.contract_base import ContractBase
from starfish.utils.did import did_to_id

logger = logging.getLogger(__name__)

CONTRACT_NAME = 'DIDRegistry'

MAX_DDO_TEXT_SIZE = 2048


class DIDRegistryContract(ContractBase):
    """

    Class representing the DIDRegistry contract.

    """

    def __init__(self):
        ContractBase.__init__(self, CONTRACT_NAME)

    def register(self, account, did, ddo_text):
        tx_hash = None
        did_id = did_to_id(did)
        if did_id:
            did_id = self._web3.toBytes(hexstr=did_id)
            if len(ddo_text) > MAX_DDO_TEXT_SIZE:
                logger.error(f'ddo test is to large {len(ddo_text)}')
            tx_hash = self.call('registerDID', (did_id, ddo_text), account)
        return tx_hash

    def get_block_number(self, did):
        block_number = None
        did_id = did_to_id(did)
        if did_id:
            did_id = self._web3.toBytes(hexstr=did_id)
            block_number = self.call('getBlockNumberUpdated', did_id)
        return block_number

    def get_value(self, did):
        block_number = self.get_block_number(did)
        if block_number:
            did_id = self._web3.toBytes(hexstr=did_to_id(did))
            event_filter = self.create_event_filter(
                'DIDRegistered',
                None,
                from_block=block_number,
                argument_filters={'_did': did_id}
            )

            if event_filter:
                for event_log in event_filter.get_all_entries():
                    event_args = event_log.args
                    return event_args['_value']
        return None



#!/usr/bin/env python
"""

    Tests for `starfish-py account`.

"""
import unittest

def test_account_load(config, accounts):
    # address only
    account = accounts[0]
    assert(account)
    assert(account.address.lower() == config.account_1['address'].lower())
    assert(account.password == config.account_1['password'])
    assert(account.keyfile == config.account_1['keyfile'])



"""

    DNetwork class


"""

import logging

from web3 import (
    HTTPProvider,
    Web3
)
from web3.gas_strategies.rpc import rpc_gas_price_strategy

from starfish.contract import ContractManager
from starfish.exceptions import StarfishInsufficientFunds

logger = logging.getLogger(__name__)

DEFAULT_PACKAGE_NAME = 'starfish.contract'

NETWORK_NAMES = {
    0: 'development',
    1: 'main',
    2: 'morden',
    3: 'ropsten',
    4: 'rinkeby',
    42: 'kovan',
    77: 'POA_Sokol',
    99: 'POA_Core',
    100: 'xDai',
    8995: 'nile',                   # Ocean Protocol Public test net
    8996: 'spree',                  # Ocean Protocol local test net
    0xcea11: 'pacific'              # Ocean Protocol Public mainnet
}

CONTRACT_LIST = {
    'Network': {
        'name': 'NetworkContract',
        'abi_filename': False,
    },
    'DIDRegistry': {
        'name': 'DIDRegistryContract',
        'abi_filename': 'DIDRegistry.development.json'
    },
    'DirectPurchase': {
        'name': 'DirectPurchaseContract'
    },
    'OceanToken': {
        'name': 'OceanTokenContract'
    },
    'Dispenser': {
        'name': 'DispenserContract'
    },
    'Provenance': {
        'name': 'ProvenanceContract',
        'abi_filename': 'Provenance.development.json'
    },

}


class DNetwork():
    def __init__(self, url):
        self._url = url
        self._web3 = None
        self._name = None
        self._contracts = {}
        self._connect(self._url)

    def get_contract(self, name):
        if name not in CONTRACT_LIST:
            raise LookupError(f'Invalid contract name: {name}')

        if name not in self._contracts:
            item = CONTRACT_LIST[name]
            self._contracts[name] = self._contract_manager.load(item['name'], self._name, item.get('abi_filename', None))
        return self._contracts[name]

    """

    Account based operations

    """
    def get_ether_balance(self, account_address):
        network_contract = self.get_contract('Network')
        return network_contract.get_balance(account_address)

    def get_token_balance(self, account):
        ocean_token_contract = self.get_contract('OceanToken')
        return ocean_token_contract.get_balance(account)

    def request_test_tokens(self, account, amount):
        dispenser_contract = self.get_contract('Dispenser')
        tx_hash = dispenser_contract.request_tokens(account, amount)
        receipt = dispenser_contract.wait_for_receipt(tx_hash)
        return receipt.status == 1

    """

    Send ether and tokens to another account

    """
    def send_ether(self, account, to_account_address, amount):
        network_contract = self.get_contract('Network')

        account_balance = self.get_ether_balance(account)
        if account_balance < amount:
            raise StarfishInsufficientFunds(f'The account has insufficient funds to send {amount} tokens')

        tx_hash = network_contract.send_ether(account, to_account_address, amount)
        receipt = network_contract.wait_for_receipt(tx_hash)
        return receipt.status == 1

    def send_token(self, account, to_account_address, amount):
        ocean_token_contract = self.get_contract('OceanToken')

        account_balance = self.get_token_balance(account)
        if account_balance < amount:
            raise StarfishInsufficientFunds(f'The account has insufficient funds to send {amount} tokens')

        tx_hash = ocean_token_contract.transfer(account, to_account_address, amount)
        receipt = ocean_token_contract.wait_for_receipt(tx_hash)
        return receipt.status == 1

    """

    Send tokens (make payment) with logging

    """
    def send_token_and_log(self, account, to_account_address, amount, reference_1=None, reference_2=None):
        ocean_token_contract = self.get_contract('OceanToken')
        direct_contract = self.get_contract('DirectPurchase')

        account_balance = self.get_token_balance(account)
        if account_balance < amount:
            raise StarfishInsufficientFunds(f'The account has insufficient funds to send {amount} tokens')

        tx_hash = ocean_token_contract.approve_transfer(
            account,
            direct_contract.address,
            amount
        )
        receipt = ocean_token_contract.wait_for_receipt(tx_hash)
        if receipt and receipt.status == 1:
            tx_hash = direct_contract.send_token_and_log(
                account,
                to_account_address,
                amount,
                reference_1,
                reference_2
            )
            receipt = direct_contract.wait_for_receipt(tx_hash)
            if receipt and receipt.status == 1:
                return True
        return False

    def is_token_sent(self, from_account_address, to_account_address, amount, reference_1=None, reference_2=None):
        direct_contract = self.get_contract('DirectPurchase')

        is_sent = direct_contract.check_is_paid(
            from_account_address,
            to_account_address,
            amount,
            reference_1,
            reference_2
        )
        return is_sent

    """

    Register Provenance

    """
    def register_provenace(self, account, asset_id):
        provenance_contract = self.get_contract('Provenance')
        tx_hash = provenance_contract.register(account, asset_id)
        receipt = provenance_contract.wait_for_receipt(tx_hash)
        return receipt.status == 1

    def get_provenace_event_list(self, asset_id):
        provenance_contract = self.get_contract('Provenance')
        return provenance_contract.get_event_list(asset_id)

    """

    Register DID with a DDO and resolve DID to a DDO

    """
    def register_did(self, account, did, ddo_text):
        did_registry_contract = self.get_contract('DIDRegistry')
        tx_hash = did_registry_contract.register(account, did, ddo_text)
        receipt = did_registry_contract.wait_for_receipt(tx_hash)
        return receipt.status == 1

    def resolve_did(self, did):
        did_registry_contract = self.get_contract('DIDRegistry')
        return did_registry_contract.get_value(did)

    @property
    def contract_names(self):
        return CONTRACT_LIST.keys()

    @property
    def url(self):
        return self._url

    @property
    def name(self):
        return self._name

    @property
    def web3(self):
        return self._web3

    @staticmethod
    def find_network_name_from_id(network_id):
        if network_id in NETWORK_NAMES:
            return NETWORK_NAMES[network_id]
        return NETWORK_NAMES[0]

    def _connect(self, url):
        self._url = url
        self._web3 = Web3(HTTPProvider(url))
        if self._web3:
            self._name = DNetwork.find_network_name_from_id(int(self._web3.net.version))
            logger.info(f'connected to the {self._name} network')
            self._web3.eth.setGasPriceStrategy(rpc_gas_price_strategy)
            self._contract_manager = ContractManager(self._web3, DEFAULT_PACKAGE_NAME)
            return True
        return False



from unittest.mock import Mock
import pytest
import secrets

from tests.unit.libs.unit_test_config import unitTestConfig
from tests.unit.libs.unit_test_network import UnitTestNetwork
from starfish import DNetwork

@pytest.fixture(scope="module")
def config():
    return unitTestConfig

@pytest.fixture(scope="module")
def network(config):
    network = UnitTestNetwork(config.network_url)
    return network

@pytest.fixture(scope='module')
def accounts(config):
    accounts = []
    for acount_info in config.accounts:
        accounts.append(Account(account_info.as_dict))
    return accounts


import pytest
import time


TOKEN_AMOUNT_TO_TRANSFER = 10

def test_ocean_taken_contract(network, accounts):

    ocean_token_contract = network.get_contract('OceanToken')
    dispenser_contract = network.get_contract('Dispenser')

    from_account = accounts[0]
    to_account = accounts[1]

    balance = ocean_token_contract.get_balance(from_account)

    tx_hash = dispenser_contract.request_tokens(from_account, TOKEN_AMOUNT_TO_TRANSFER * 2)
    receipt = dispenser_contract.wait_for_receipt(tx_hash)

    # give enougth time for the block chain to go to the next block and mine this dispenser request
    time.sleep(1)

    from_balance = ocean_token_contract.get_balance(from_account)
    to_balance = ocean_token_contract.get_balance(to_account)

#    tx_hash = ocean_token_contract.approve_transfer(from_account, ocean_token_contract.address, TOKEN_AMOUNT_TO_TRANSFER)
#    receipt = ocean_token_contract.wait_for_receipt(tx_hash)
    """

        *********************************************************************
        This does not work on the github test suite, but works on the desktop
        *********************************************************************

    tx_hash = ocean_token_contract.transfer(from_account, to_account.address, TOKEN_AMOUNT_TO_TRANSFER)
    receipt = ocean_token_contract.wait_for_receipt(tx_hash)

    new_from_balance = ocean_token_contract.get_balance(from_account)
    new_to_balance = ocean_token_contract.get_balance(to_account)

    assert(from_balance - TOKEN_AMOUNT_TO_TRANSFER == new_from_balance)
    assert(to_balance + TOKEN_AMOUNT_TO_TRANSFER == new_to_balance)

    """


import pytest
import secrets
import time

TOKEN_AMOUNT_TO_TRANSFER = 10


def test_direct_purchase(network, accounts):
    """

    Transfer funds from the 'buy' account -> 'sell' account

    """
    direct_contract = network.get_contract('DirectPurchase')
    ocean_token_contract = network.get_contract('OceanToken')
    dispenser_contract = network.get_contract('Dispenser')

    from_account = accounts[0]
    to_account = accounts[1]
    ref_1 = secrets.token_hex(32)
    ref_2 = secrets.token_hex(32)

    tx_hash = dispenser_contract.request_tokens(from_account, TOKEN_AMOUNT_TO_TRANSFER)
    receipt = dispenser_contract.wait_for_receipt(tx_hash)
    # print('request tokens receipt', receipt)
    assert(receipt.status == 1)

    from_balance = ocean_token_contract.get_balance(from_account)
    to_balance = ocean_token_contract.get_balance(to_account)

    # ocean_token_contract.unlockAccount(from_account)

    tx_hash = ocean_token_contract.approve_transfer(
        from_account,
        direct_contract.address,
        TOKEN_AMOUNT_TO_TRANSFER
    )
    receipt = ocean_token_contract.wait_for_receipt(tx_hash)
    # print('approve transfer receipt', receipt)
    assert(receipt.status == 1)


    tx_hash = direct_contract.send_token_and_log(
        from_account,
        to_account.address,
        TOKEN_AMOUNT_TO_TRANSFER,
        ref_1,
        ref_2
    )
    receipt = dispenser_contract.wait_for_receipt(tx_hash)
    # print('send token and log receipt', receipt)
    assert(receipt.status == 1)

    new_from_balance = ocean_token_contract.get_balance(from_account)
    new_to_balance = ocean_token_contract.get_balance(to_account)

    assert(from_balance - TOKEN_AMOUNT_TO_TRANSFER == new_from_balance)
    assert(to_balance + TOKEN_AMOUNT_TO_TRANSFER == new_to_balance)

    is_sent = direct_contract.check_is_paid(
        from_account.address,
        to_account.address,
        TOKEN_AMOUNT_TO_TRANSFER,
        ref_1
    )
    assert(is_sent)

    is_sent = direct_contract.check_is_paid(
        from_account.address,
        to_account.address,
        TOKEN_AMOUNT_TO_TRANSFER,
        ref_1,
        ref_2
    )
    assert(is_sent)




import os
import re

from configparser import (
    ConfigParser,
    ExtendedInterpolation,
)


class IntegrationTestConfig():
    def __init__(self, filename):

        config = ConfigParser(interpolation=ExtendedInterpolation())
        config.read(filename)

        if 'NETWORK_URL' in os.environ:
            config.set('test', 'network_url', os.environ['NETWORK_URL'])

        self.network_url = config.get('network', 'url')

        self.account_1 = {
            'address': config.get('account 1', 'address'),
            'password': config.get('account 1', 'password'),
            'keyfile': config.get('account 1', 'keyfile'),
        }
        self.account_2 = {
            'address': config.get('account 2', 'address'),
            'password': config.get('account 2', 'password'),
            'keyfile': config.get('account 2', 'keyfile'),
        }

        # load in list of agents

        self.agent_list = {}
        for section_name in config.sections():
            match = re.match(r'(\w+)\s+agent', section_name, re.IGNORECASE)
            if match:
                agent_name = match.groups(1)[0]
                self.agent_list[agent_name] = {
                    'url': config.get(section_name, 'url', fallback=None),
                    'did': config.get(section_name, 'did', fallback=None),
                    'username': config.get(section_name, 'username'),
                    'password': config.get(section_name, 'password')
                }





"""

    DIDRegistry Contract

"""

import datetime
import logging
from eth_utils import remove_0x_prefix

from starfish.contract.contract_base import ContractBase

logger = logging.getLogger(__name__)

CONTRACT_NAME = 'Provenance'


class ProvenanceContract(ContractBase):
    """

    Class representing the DIDRegistry contract.

    """

    def __init__(self):
        ContractBase.__init__(self, CONTRACT_NAME)

    def register(self, account, asset_id):
        asset_id_bytes = self._web3.toBytes(hexstr=asset_id)
        tx_hash = self.call('registerAsset', (asset_id_bytes), account)
        return tx_hash

    def get_block_number(self, asset_id):
        asset_id_bytes = self._web3.toBytes(hexstr=asset_id)
        block_number = self.call('getBlockNumber', asset_id_bytes)
        return block_number

    def get_event_list(self, asset_id):
        result = []
        from_block_number = self.get_block_number(asset_id)
        if from_block_number:
            asset_id_bytes = self._web3.toBytes(hexstr=asset_id)
            event_filter = self.create_event_filter(
                'AssetRegistered',
                None,
                from_block=from_block_number,
                argument_filters={'_assetID': asset_id_bytes}
            )

            if event_filter:
                for event_log in event_filter.get_all_entries():
                    event_args = event_log.args
                    event_asset_id = remove_0x_prefix(self._web3.toHex(event_args['_assetID']))
                    if remove_0x_prefix(event_asset_id) == remove_0x_prefix(asset_id):
                        item = {
                            'asset_id': event_asset_id,
                            'account': event_args['_user'],
                            'timestamp': datetime.datetime.fromtimestamp(event_args['_timestamp'])
                        }
                        result.append(item)
        return result



"""
    test_08_asset_registration

    As a developer working with Ocean,
    I need a way to register a new asset with Ocean

"""

import secrets
import json

from starfish.asset import DataAsset
from starfish.agent import RemoteAgent


def test_08_asset_registration(resources, remote_agent):
    test_data = secrets.token_hex(1024)
    asset_data_1 = DataAsset.create('TestAsset', test_data)
    asset_1 = remote_agent.register_asset(asset_data_1)
    assert(asset_1)

    asset_data_2 = DataAsset.create('TestAsset', test_data)
    asset_2 = remote_agent.register_asset(asset_data_2)
    assert(asset_2)

    assert(asset_1.asset_id == asset_2.asset_id)



"""

Memory Agent class to provide basic functionality for Ocean Agents

"""

import json
import re
import secrets

from starfish.agent import AgentBase
from starfish.ddo.ddo import DDO
from starfish.listing import Listing
from starfish.purchase import Purchase
from starfish.utils.crypto_hash import hash_sha3_256
from starfish.utils.did import (
    did_generate_random,
    did_parse,
    did_to_asset_id
)


class MemoryAgent(AgentBase):
    """

    Memory Agent class allows to register, list, purchase and consume assets.

    :param ocean: Ocean object that is being used.
    :type ocean: :class:`.Ocean`

    """

    def __init__(self, network, ddo=None):

        if ddo is None:
            did = did_generate_random()
            ddo = DDO(did)

        AgentBase.__init__(self, network, ddo=ddo)

        self._memory = {
            'listing': {},
            'asset': {},
            'purchase': {}
        }

    def register_asset(self, asset):
        """

        Register a memory asset with the ocean network.

        :param object metadata: asset object to register for this asset.

        :return: A :class:`.AssetBase` object that has been registered, if failure then return None.
        :type: :class:`.AssetBase` class

        """

        asset_id = hash_sha3_256(asset.metadata_text)
        did = f'{self._ddo.did}/{asset_id}'

        self._memory['asset'][did] = asset
        asset.set_did(did)
        return asset

    def create_listing(self, listing_data, asset_did):
        """

        Create a listing on the market place for this asset

        :param dict listing_data:  Listing inforamiton to give for this asset
        :param str asset_did: asset DID to assign to this listing
        :return: A new :class:`.Listing` object that has been registered, if failure then return None.
        :type: :class:`.Listing` class

        """
        listing = None
        if listing_data:
            listing_id = did_to_asset_id(asset_did)
            listing_data['listing_id'] = listing_id,
            listing_data['asset_did'] = asset_did
            listing = Listing(self, listing_id, asset_did, listing_data)
            self._memory['listing'][listing_id] = listing
        return listing

    def update_listing(self, listing):
        """

        Update the listing to the agent server.

        :param listing: Listing object to update
        :type listing: :class:`.Listing` class

        """
        self._memory['listing'][listing.listing_id] = listing

    def validate_asset(self, asset):
        """

        Validate an asset

        :param asset: Asset to validate.
        :return: True if the asset is valid
        """
        return asset is not None

    def get_listing(self, listing_id):
        """

        Return an listing from the given listing_id.

        :param str listing_id: Id of the listing.

        :return: a registered listing given a Id of the listing
        :type: :class:`.Listing` class

        """
        listing = None
        if listing_id in self._memory['listing']:
            listing = self._memory['listing'][listing_id]
        return listing

    def search_listings(self, text, sort=None, offset=100, page=0):
        """

        Search for listings with the givien 'text'

        :param str text: Text to search all listing data for.
        :param sort: sort the results ( defaults: None, no sort).
        :type sort: str or None
        :param int offset: Return the result from with the maximum record count ( defaults: 100 ).
        :param int page: Returns the page number based on the offset.

        :return: a list of listing objects found using the search.
        :type: :class:`.Listing` objects

        For example::

            # return the 300 -> 399 records in the search for the text 'weather' in the metadata.
            my_result = agent.search_registered_assets('weather', None, 100, 3)

        """
        listing_id_list = []
        for listing_id, listing in self._memory['listing'].items():
            if re.search(text, json.dumps(listing.data)):
                listing_id_list.append(listing.listing_id)

        return listing_id_list

    def purchase_asset(self, listing, account):
        """

        Purchase an asset using it's listing and an account.

        :param listing: Listing to use for the purchase.
        :type listing: :class:`.Listing`
        :param account: Ocean account to purchase the asset.
        :type account: :class:`.Account` object to use for registration.

        """
        purchase = None

        purchase_id = secrets.token_hex(64)
        if purchase_id:
            purchase = Purchase(self, listing, purchase_id, account)
            self._memory['purchase'][purchase_id] = (purchase, account.address)

        return purchase

    def purchase_wait_for_completion(self, asset, account,  purchase_id, timeoutSeconds):
        """

            Wait for completion of the purchase

            TODO: issues here...
            + No method as yet to pass back paramaters and values during the purchase process
            + We assume that the following templates below will always be used.

        """
        return True

    def is_access_granted_for_asset(self, asset, account, purchase_id=None):
        """

        Check to see if the account and purchase_id have access to the assed data.


        :param asset: Asset to check for access.
        :type asset: :class:`.Asset` object
        :param account: Ocean account to purchase the asset.
        :type account: :class:`.Account` object to use for registration.
        :param str purchase_id: purchase id that was used to purchase the asset.

        :return: True if the asset can be accessed and consumed.
        :type: boolean
        """

        if purchase_id in self._memory['purchase']:
            purchase, account_address = self._memory['purchase'][purchase_id]
            return purchase and account.is_address_equal(account_address)

        return False

    def get_asset_purchase_ids(self, asset):
        """

        Returns as list of purchase id's that have been used for this asset

        :param asset: DataAsset to return purchase details.
        :type asset: :class:`.DataAsset` object

        :return: list of purchase ids
        :type: list

        """
        return []

    def consume_asset(self, listing, account, purchase_id):
        """
        Consume the asset and download the data. The actual payment to the asset
        provider will be made at this point.

        :param listing: Listing that was used to make the purchase.
        :type listing: :class:`.Listing`
        :param str purchase_id: purchase id that was used to purchase the asset.
        :param account: Ocean account that was used to purchase the asset.
        :type account: :class:`.Account` object to use for registration.

        :return: True if the asset has been consumed and downloaded
        :type: boolean

        """
        return purchase_id in self._memory['purchase']

    @staticmethod
    def is_did_valid(did):
        """
        Checks to see if the DID string is a valid DID for this type of Asset.
        This method only checks the syntax of the DID, it does not resolve the DID
        to see if it is assigned to a valid Asset.

        :param str did: DID string to check to see if it is in a valid format.

        :return: True if the DID is in the format 'did:dep:xxxxx'
        :type: boolean
        """
        data = did_parse(did)
        return not data['path']



"""

Utils: Data Bundle.

Allow to split up data into 'chunks' and each chunk is an asset, and all of the assets are held in a bundle asset


"""

import math
import os
import re

from starfish.asset import (
    BundleAsset,
    DataAsset
)

# WARNING currently surfer cannot support > 6mb in asset data size
DEFAULT_CHUNK_SIZE = '6mb'


def decode_readable_size(text, base_size=1024):
    sizes = {
        r'(\d)\s?b.?': 0,       # bytes
        r'([\d\.]+)\s?k.?': 1,   # kilobyte
        r'([\d\.]+)\s?m.?': 2,   # megabyte
        r'([\d\.]+)\s?g.?': 3,   # gigabyte
        r'([\d\.]+)\s?t.?': 4,   # terabyte
        r'([\d\.]+)\s?p.?': 5,   # petabyte
        r'([\d\.]+)\s?e.?': 6,   # exabyte
        r'([\d\.]+)\s?z.?': 7,   # zettabyte
        r'([\d\.]+)\s?y.?': 8,   # yottabyte

    }
    for regexp, factor in sizes.items():
        match = re.match(regexp, text, re.IGNORECASE)
        if match:
            return int(float(match.group(1)) * math.pow(base_size, factor))
    return None


def register_upload_data(remote_agent, name, data_stream, chunk_size_value=None):

    if chunk_size_value is None:
        chunk_size_value = DEFAULT_CHUNK_SIZE

    chunk_size = chunk_size_value
    if isinstance(chunk_size_value, str):
        chunk_size = decode_readable_size(chunk_size_value)
    if not isinstance(chunk_size, int):
        raise ValueError(f'invalid chunk size "{chunk_size_value}", must be int or str')

    bundle_asset = BundleAsset.create(name)
    index = 0
    asset = None
    while True:
        data = data_stream.read(chunk_size)
        if data:
            asset_name = f'{name}:{index}'
            data_asset = DataAsset.create(asset_name, data)
            asset = remote_agent.register_asset(data_asset)
            remote_agent.upload_asset(asset)
            bundle_asset.add(asset_name, asset)
            index += 1
        else:
            break
    if index > 0:
        asset = remote_agent.register_asset(bundle_asset)
    return asset


def register_upload_bundle_file(remote_agent, filename, chunk_size=None):
    if not os.path.exists(filename):
        raise FileNotFoundError(f'Cannot find file {filename}')
    bundle_asset = None
    with open(filename, 'rb') as fp:
        name = f'file: {os.path.basename(filename)}'
        bundle_asset = register_upload_data(remote_agent, name, fp, chunk_size)
    return bundle_asset


def download_bundle_data(remote_agent, bundle_asset, data_stream):
    size = 0
    for name, asset in bundle_asset:
        store_asset = remote_agent.download_asset(asset.asset_id)
        data_stream.write(store_asset.data)
        size += len(store_asset.data)
    return size


def download_bundle_file(remote_agent, bundle_asset, filename):
    if not bundle_asset.is_bundle:
        raise TypeError(f'asset type {bundle_asset.type} is not a bundle asset')
    size = 0
    with open(filename, 'wb') as fp:
        size = download_bundle_data(remote_agent, bundle_asset, fp)
    return size



"""

Unit test AgentManager

"""
import pytest

from starfish.agent import AgentManager

def test_agent_manager_add(network):
    manager = AgentManager(network)

    manager.add('test', 'http://test.com', did=None, username='user_test', password='password_test')
    assert(manager.items['test'])

    with pytest.raises(ValueError):
        manager.add('failed')



"""
    Test ocean class

"""

import datetime
import pathlib
import json
import logging
import time
import secrets
from web3 import Web3

from starfish.agent import MemoryAgent
from starfish.asset import DataAsset


def register_asset_for_sale(agent, resources, account):

    asset = DataAsset.create('TestAsset', secrets.token_hex(256))
    asset = agent.register_asset(asset)
    listing = agent.create_listing(resources.listing_data, asset.did)
    assert listing
    assert listing.asset_did
    return listing

def test_asset(network, resources, config, accounts):

    # create an ocean object

    publisher_account = accounts[0]
    purchaser_account = accounts[1]

    agent = MemoryAgent(network)
    assert agent


    listing = register_asset_for_sale(agent, resources, publisher_account)
    assert listing
    assert publisher_account

    listing_id = listing.listing_id
    # start to test getting the asset from storage
    listing = agent.get_listing(listing_id)
    assert listing
    assert listing.listing_id == listing_id


    # test purchase an asset
    purchase_asset = listing.purchase(purchaser_account)
    assert purchase_asset


    assert(purchase_asset.is_purchased)
    assert(purchase_asset.wait_for_completion())
    assert(purchase_asset.is_completed)
    assert(purchase_asset.is_purchase_valid)

    assert(purchase_asset.consume_asset)



def test_search_listing(network, resources, config, accounts):

    publisher_account = accounts[0]
    purchaser_account = accounts[1]

    agent = MemoryAgent(network)

    listing = register_asset_for_sale(agent, resources, publisher_account)
    assert listing
    assert publisher_account

    # choose a word from the description field
    text = resources.listing_data['author']
    words = text.split(' ')
    word = words[0]

    # should return at least 1 or more assets
    logging.info(f'search word is {word}')
    searchResult = agent.search_listings(word)
    assert searchResult

    assert(len(searchResult) > 0)



"""
    test_20_invoke_service_free

    As a developer working with Ocean,
    I wish to invoke a free service available on the Ocean ecosystem and obtain the results as a new Ocean Asset

"""

import secrets
import logging
import json
import time
import re
import requests

from starfish.asset import OperationAsset
from starfish.job import Job
from starfish.utils.did import did_to_asset_id

INVOKE_TOKENIZE_TEXT_NAME = "Tokenize Text"
INVOKE_INCREMENT_NAME = "Increment"

TEST_TEXT = 'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Vestibulum eu congue odio, vel congue sapien. Morbi ac purus ornare, volutpat elit a, lacinia odio. Integer tempor tellus eget iaculis lacinia. Curabitur aliquam, dui vel vestibulum rhoncus, enim metus interdum enim, in sagittis massa est vel velit. Nunc venenatis commodo libero, vitae elementum nulla ultricies id. Aliquam erat volutpat. Cras eu pretium lacus, quis facilisis mauris. Duis sem quam, blandit id tempor in, porttitor at neque. Cras ut blandit risus. Maecenas vitae sodales neque, eu ultricies nibh.'


def load_operation_asset(remote_agent, invokable_list, name):

    # cannot find {name} in invokable list
    assert(name in invokable_list)

    asset_did = invokable_list[name]

    assert(asset_did)

    asset_id = did_to_asset_id(asset_did)

    # no operation asset found
    assert(asset_id)

    asset = remote_agent.get_asset(asset_id)

    # cannot load asset
    assert(asset)

    # asset returned is not an operation asset
    assert(isinstance(asset, OperationAsset))

    # invalid asset type
    assert(asset.metadata['type'] == 'operation')

    return asset

def assert_tokenize_text(tokens, text):
    values = re.split(r'\W+', text.lower())
    values.pop()
    assert(tokens == values)


def test_20_tokenize_text_sync(remote_agent, invokable_list):

    asset = load_operation_asset(remote_agent, invokable_list, INVOKE_TOKENIZE_TEXT_NAME)
    inputs = {
        'text': TEST_TEXT
    }
    response = remote_agent.invoke(asset, inputs)
    assert(response)
    assert(response['outputs'])
    assert(response['status'])
    assert(response['status'] == 'succeeded')
    assert_tokenize_text(response['outputs']['tokens'], TEST_TEXT)

def test_20_tokenize_text_async(remote_agent, invokable_list):

    asset = load_operation_asset(remote_agent, invokable_list, INVOKE_TOKENIZE_TEXT_NAME)

    inputs = {
        'text': TEST_TEXT
    }

    response = remote_agent.invoke(asset, inputs, True)
    assert(response)
    assert(response['job-id'])

    job_id = response['job-id']
    assert(isinstance(job_id, str))

    # wait for job to complete
    time.sleep(1)
    # test get_job
    job = remote_agent.get_job(job_id)
    assert(job)
    assert(isinstance(job, Job))

    # test wait_for_job
    job = remote_agent.job_wait_for_completion(job_id)
    assert(job)
    assert(isinstance(job, Job))
    assert(job.status == 'succeeded')
    assert(job)
    assert_tokenize_text(job.outputs['tokens'], TEST_TEXT)



def test_20_increment_sync(remote_agent, invokable_list):


    value = time.time()
    asset = load_operation_asset(remote_agent, invokable_list, INVOKE_INCREMENT_NAME)

    inputs = {
        'n': value
    }

    response = remote_agent.invoke(asset, inputs)
    assert(response)
    assert(response['status'])
    assert(response['status'] == 'succeeded')
    assert(response['outputs'])
    assert(float(response['outputs']['n']) - 1 == value)

def test_20_increment_async(remote_agent, invokable_list):

    value = time.time()
    asset = load_operation_asset(remote_agent, invokable_list, INVOKE_INCREMENT_NAME)


    inputs = {
        'n': value
    }

    response = remote_agent.invoke(asset, inputs, True)
    assert(response)
    assert(response['job-id'])

    job_id = response['job-id']
    assert(isinstance(job_id, str))

    # wait for job to complete
    time.sleep(1)
    # test get_job
    job = remote_agent.get_job(job_id)
    assert(job)
    assert(isinstance(job, Job))

    # test wait_for_job
    job = remote_agent.job_wait_for_completion(job_id)
    assert(job)
    assert(isinstance(job, Job))
    assert(job.status == 'succeeded')
    assert(job)
    assert(float(job.outputs['n']) - 1 == value)



"""

    Provenance module

"""

import secrets

PROVENANCE_DEP = 'dep'
PROVENANCE_ACTIVITY_TYPE_PUBLISH = 'publish'
PROVENANCE_ACTIVITY_TYPE_PUBLISH = 'import'
PROVENANCE_ACTIVITY_TYPE_OPERATION = 'operation'

PROVENANCE_AGENT_TYPE_ACCOUNT = 'account'


def create_publish(agent_id, activity_id=None):
    if activity_id is None:
        activity_id = secrets.token_hex(32)

    return {
        'prefix': create_prefix(),
        'activity': create_activity(activity_id, PROVENANCE_ACTIVITY_TYPE_PUBLISH),
        'entity': add_asset_entity('this'),
        'agent': create_agent(agent_id, PROVENANCE_AGENT_TYPE_ACCOUNT),
        'wasAssociatedWith': create_associated_with(agent_id, activity_id),
        'wasGeneratedBy': create_generated_by(activity_id)
    }


def create_invoke(activity_id, agent_id, asset_list, inputs_text, outputs_text):
    entities = add_asset_entity('this')
    for asset in asset_list:
        entities = add_asset_entity(asset, entities)

    dependencies = create_dependencies(inputs_text, outputs_text)
    return {
        'prefix': create_prefix(),
        'activity': create_activity(activity_id, PROVENANCE_ACTIVITY_TYPE_OPERATION, dependencies),
        'entity': entities,
        'agent': create_agent(agent_id, PROVENANCE_AGENT_TYPE_ACCOUNT),
        'wasAssociatedWith': create_associated_with(agent_id, activity_id),
        'wasDerivedFrom': create_derived_from(asset_list),
        'wasGeneratedBy': create_generated_by(activity_id)
    }


def create_prefix():
    return {
        'xsd': 'http://www.w3.org/2001/XMLSchema#',
        'prov': 'http://www.w3.org/ns/prov#',
        PROVENANCE_DEP: 'http://dex.sg'
    }


def add_asset_entity(asset_id, entities=None):
    if entities is None:
        entities = {}
    entities[f'{PROVENANCE_DEP}:{asset_id}'] = {
        'prov:type': {
            '$': f'{PROVENANCE_DEP}:asset',
            'type': 'xsd:string'
        }
    }
    return entities


def create_activity(activity_id, activity_type, entries=None):
    items = {
        'prov:type': {
            '$': f'{PROVENANCE_DEP}:{activity_type}',
            'type': 'xsd:string'
        }
    }
    if entries:
        for name, value in entries.items():
            items[name] = value

    return {
        f'{PROVENANCE_DEP}:{activity_id}': items
    }


def create_agent(agent_id, agent_type):
    return {
        f'{PROVENANCE_DEP}:{agent_id}': {
            'prov:type': {
                '$': f'{PROVENANCE_DEP}:{agent_type}',
                'type': 'xsd:string'
            }
        }
    }


def create_associated_with(agent_id, activity_id):
    random_id = secrets.token_hex(32)
    return {
        f'_:{random_id}': {
            'prov:agent': agent_id,
            'prov:activity': activity_id
        }
    }


def create_generated_by(activity_id):
    entity_id = f'{PROVENANCE_DEP}:this'
    random_id = secrets.token_hex(32)
    return {
        f'_:{random_id}': {
            'prov:entity': entity_id,
            'prov:activity': activity_id
        }
    }


def create_derived_from(asset_list):
    entities = {}
    for asset in asset_list:
        random_id = secrets.token_hex(32)
        entities[f'_:{random_id}'] = {
            'prov:usedEntity': asset.asset_id,
            'prov:generatedEntity': f'{PROVENANCE_DEP}:this'
        }
    return entities


def create_dependencies(inputs_text, outputs_text):
    return {
        f'{PROVENANCE_DEP}:outputs': {
            '$': outputs_text,
            'type': 'xsd:string'
        },
        f'{PROVENANCE_DEP}:inputs': {
            '$': inputs_text,
            'type': 'xsd:string'
        }
    }



import pytest
import secrets
import time

from starfish.utils.did import did_generate_random
from starfish.agent import RemoteAgent
from starfish.agent.services import Services

ASSET_REGISTER_COUNT = 2

def test_provenance_contract(network, accounts):
    """

    Register and find a provenance record on the block chain for a given asset_id

    """
    provenance_contract = network.get_contract('Provenance')

    asset_id = secrets.token_hex(32)

    register_account = accounts[0]

    register_block_number = 0
    for counter in range(0, ASSET_REGISTER_COUNT):
        tx_hash = provenance_contract.register(register_account, asset_id)
        receipt = provenance_contract.wait_for_receipt(tx_hash)
        if register_block_number == 0:
            register_block_number = receipt.blockNumber

    block_number = provenance_contract.get_block_number(asset_id)
    assert(register_block_number == block_number)

    event_list = provenance_contract.get_event_list(asset_id)
    assert(len(event_list) == ASSET_REGISTER_COUNT)
    for event in event_list:
        assert(event['asset_id'] == asset_id)



"""
    test_asset_base

"""

import secrets
import logging
import json

from starfish.asset import AssetBase
from starfish.utils.did import (
    did_to_id,
    id_to_did,
    did_to_asset_id
)

ASSET_METADATA = {
    'name': 'Asset',
    'type': 'asset',
}

TEST_DID = id_to_did(secrets.token_hex(32))

def test_init(metadata):
    asset = AssetBase(ASSET_METADATA)
    assert(asset)
    assert(isinstance(asset, AssetBase))

def test_metadata():
    asset = AssetBase(ASSET_METADATA)
    assert(asset)
    assert(asset.metadata == ASSET_METADATA)

def test_data():
    asset = AssetBase(ASSET_METADATA, TEST_DID)
    assert(asset)
    assert(asset.metadata == ASSET_METADATA)
    assert(asset.did == TEST_DID)


def test_is_asset_type():
    asset = AssetBase(ASSET_METADATA)
    assert(asset)
    assert(asset.is_asset_type('asset'))
    assert(not asset.is_asset_type('bad asset type'))

def test_asset_id():
    asset = AssetBase(ASSET_METADATA, TEST_DID)
    assert(asset)
    asset_id = did_to_asset_id(asset.did)
    print(asset_id, asset.asset_id)

    assert(asset.asset_id == asset_id)

    did_id = secrets.token_hex(32)
    path_did = id_to_did(did_id)
    asset_id = did_to_asset_id(f'{path_did}/{asset_id}')

    asset = AssetBase(ASSET_METADATA, f'{path_did}/{asset_id}')
    assert(asset)
    assert(asset.asset_id == asset_id)



"""

Test utils.data_bundle module



"""
import math

from starfish.utils.data_bundle import decode_readable_size

ONE_TB = math.pow(1024, 4)
ONE_GB = math.pow(1024, 3)
ONE_MB = math.pow(1024, 2)
ONE_KB = 1024


def test_decode_readable_size():

    test_values = {
        '1mb': ONE_MB,
        '1 mb': ONE_MB,
        '125mb': ONE_MB * 125,
        '2m': ONE_MB * 2,
        '1kb': ONE_KB,
        '20.2kb': ONE_KB * 20.2,
        '4k': ONE_KB * 4,
        '502GB': ONE_GB * 502,
        '6G': ONE_GB * 6,
        '22T': ONE_TB * 22,
        '1 Tb': ONE_TB
    }
    for text, actual_value in test_values.items():
        value = decode_readable_size(text)
        assert(value, actual_value)




"""

    Contract Module


"""

from starfish.contract.contract_manager import ContractManager                      # noqa: F401
from starfish.contract.did_registry_contract import DIDRegistryContract             # noqa: F401
from starfish.contract.direct_purchase_contract import DirectPurchaseContract       # noqa: F401
from starfish.contract.dispenser_contract import DispenserContract                  # noqa: F401
from starfish.contract.network_contract import NetworkContract                      # noqa: F401
from starfish.contract.ocean_token_contract import OceanTokenContract               # noqa: F401
from starfish.contract.provenance_contract import ProvenanceContract                # noqa: F401



"""
    test_06_agent_endpoint_query


    As a developer working with Ocean,
    I need to obtain service endpoints for an arbitrary Agent in the Ocean ecosystem

"""

import re

from starfish.agent import RemoteAgent


def test_06_agent_endpoint_query(remote_agent):

    endpoint = remote_agent.get_endpoint('meta')
    assert(re.search('/meta', endpoint))
    endpoint = remote_agent.get_endpoint('storage')
    assert(re.search('/assets', endpoint))
    endpoint = remote_agent.get_endpoint('invoke')
    assert(re.search('/invoke', endpoint))
    endpoint = remote_agent.get_endpoint('market')
    assert(re.search('/market', endpoint))
    endpoint = remote_agent.get_endpoint('trust')
    assert(re.search('/trust', endpoint))
    endpoint = remote_agent.get_endpoint('auth')
    assert(re.search('/auth', endpoint))



"""


    Contract Manager

    loads contracts


"""

import importlib
import inspect
import json
import logging
import os

from starfish.contract.contract_base import ContractBase

logger = logging.getLogger(__name__)


class ContractManager:

    def __init__(self, web3, default_package_name):
        self._web3 = web3
        self._default_package_name = default_package_name

    def load(self, name, network_name, abi_filename=None, package_name=None):
        if package_name is None:
            package_name = self._default_package_name
        if package_name is None:
            raise ValueError(f'You need to provide a package name for the contract classes')

        package_module = importlib.import_module(package_name)
        for item in inspect.getmembers(package_module, inspect.ismodule):
            class_def = ContractManager._find_class_in_module(name, item[1])
            if class_def:
                return self.create_contract(name, network_name, class_def, abi_filename)
        return None

    def create_contract(self, name, network_name, class_def, abi_filename=None):
        contract_object = class_def()
        contract_name = contract_object.name
        if abi_filename is None:
            abi_filename = f'{contract_name}.{network_name}.json'
        if abi_filename:
            abi_filename_path = ContractManager.find_abi_filename(abi_filename)
            if abi_filename_path:
                contract_info = ContractManager.load_abi_file(abi_filename_path)
                contract_object.load(self.web3, abi=contract_info['abi'], address=contract_info['address'])
            else:
                raise FileNotFoundError(f'Cannot find artifact file for contract {contract_name} {abi_filename}')
        else:
            # load in an dummy contract with no abi or address
            contract_object.load(self.web3)

        return contract_object

    @property
    def web3(self):
        return self._web3

    @staticmethod
    def _find_class_in_module(class_name, contract_module):
        for name, obj in inspect.getmembers(contract_module, inspect.isclass):
            if issubclass(obj, ContractBase) \
               and name != 'ContractBase' \
               and name == class_name:
                return obj
        return None

    @staticmethod
    def find_abi_filename(filename):
        test_file = os.path.join('artifacts', filename)
        if os.path.exists(test_file):
            return test_file
        return None

    @staticmethod
    def load_abi_file(filename):
        if filename and os.path.exists(filename):
            with open(filename, 'r') as fp:
                return json.load(fp)



"""

    Ocean Token Contract

"""

from .contract_base import ContractBase

CONTRACT_NAME = 'OceanToken'


class OceanTokenContract(ContractBase):
    """Class representing the Ocean Token contract."""

    def __init__(self):
        ContractBase.__init__(self, CONTRACT_NAME)

    def get_balance(self, account_address):
        address = self.get_account_address(account_address)
        amount_wei = self.call('balanceOf', address)
        return self.to_ether(amount_wei)

    def approve_transfer(self, account, to_account_address, amount):
        to_address = self.get_account_address(to_account_address)
        amount_wei = self.to_wei(amount)
        return self.call('approve', (to_address, amount_wei), account)

    def transfer(self, account, to_account_address, amount):
        to_address = self.get_account_address(to_account_address)
        amount_wei = self.to_wei(amount)
        return self.call('transfer', (to_address, amount_wei), account)

    def total_supply(self):
        amount_wei = self.call('totalSupply')
        return self.to_ether(amount_wei)



"""
    test_02_asset_identity


    As a developer working with Ocean,
    I need a stable identifier (Asset ID) for an arbitrary asset in the Ocean ecosystem

"""

import secrets
from starfish.asset import DataAsset
from starfish.agent import RemoteAgent


def test_02_asset_register(remote_agent, resources):
    testData = secrets.token_bytes(1024)
    asset1 = DataAsset.create('TestAsset1', testData)
    asset2 = DataAsset.create('TestAsset2', testData)
    asset = remote_agent.register_asset(asset2)
    assert(asset.data == asset2.data)
    assert(asset.asset_id == asset2.asset_id)


def test_02_asset_upload(remote_agent, resources):
    testData = secrets.token_bytes(1024)
    asset_data = DataAsset.create('TestAsset', testData)
    asset = remote_agent.register_asset(asset_data)
    remote_agent.upload_asset(asset)



"""

Test provenance


"""
import json
import re
import secrets

from starfish.provenance import create_publish, create_invoke
from starfish.asset import DataAsset, OperationAsset
from starfish.metadata import METADATA_PROVENANCE_NAME

OPERATION_METADATA = {
    'name': 'operation',
    'type': 'operation',
    'operation': {
        'modes': ['sync', 'async', 'test'],
    },
}


PUBLISH_STANDARD = {
    'prefix':{
        'xsd': 'http://www.w3.org/2001/XMLSchema#',
        'prov': 'http://www.w3.org/ns/prov#',
        'dep': 'http://dex.sg'
    },
    'activity': {
        'dep:<activity_id>': {
            'prov:type': {
                '$': 'dep:import',
                'type': 'xsd:string'
            }
        }
    },
    'entity': {
        'dep:this': {
            'prov:type': {
                '$': 'dep:asset',
                'type': 'xsd:string'
            }
        }
    },
    'agent': {
        'dep:<agent_id>': {
            'prov:type': {
                '$': 'dep:account',
                'type': 'xsd:string'
            }
        }
    },
    'wasAssociatedWith': {
        '_:<random_id>': {
            'prov:agent': '<agent_id>',
            'prov:activity': '<activity_id>'
        }
    },
    'wasGeneratedBy': {
        '_:<random_id>': {
            'prov:entity': 'dep:this',
            'prov:activity': '<activity_id>'
        }
    }
}

def assert_walk_dict(standard, result):
    if isinstance(result, dict):
        assert(isinstance(standard, dict))
        for name, values in standard.items():
            assert(name in result)
            if isinstance(result[name], dict):
                assert_walk_dict(standard[name], result[name])
            else:
                assert(standard[name] == result[name])

def test_provenance_publish():

    agent_id = secrets.token_hex(32)
    result = create_publish(agent_id)
    assert(isinstance(result, dict))

    agent_dep = f'dep:{agent_id}'
    PUBLISH_STANDARD['agent'][agent_dep] =  PUBLISH_STANDARD['agent']['dep:<agent_id>']
    del PUBLISH_STANDARD['agent']['dep:<agent_id>']

    activity_dep = list(result['activity'].keys())[0]
    activity_id = re.sub(r'dep:', '', activity_dep)
    PUBLISH_STANDARD['activity'][activity_dep] =  PUBLISH_STANDARD['activity']['dep:<activity_id>']
    del PUBLISH_STANDARD['activity']['dep:<activity_id>']

    id_dep = list(result['wasAssociatedWith'].keys())[0]
    PUBLISH_STANDARD['wasAssociatedWith'][id_dep] =  PUBLISH_STANDARD['wasAssociatedWith']['_:<random_id>']
    del PUBLISH_STANDARD['wasAssociatedWith']['_:<random_id>']


    PUBLISH_STANDARD['wasAssociatedWith'][id_dep]['prov:agent'] = agent_id
    PUBLISH_STANDARD['wasAssociatedWith'][id_dep]['prov:activity'] = activity_id

    id_dep = list(result['wasGeneratedBy'].keys())[0]
    PUBLISH_STANDARD['wasGeneratedBy'][id_dep] =  PUBLISH_STANDARD['wasGeneratedBy']['_:<random_id>']
    del PUBLISH_STANDARD['wasGeneratedBy']['_:<random_id>']

    PUBLISH_STANDARD['wasGeneratedBy'][id_dep]['prov:activity'] = activity_id

    assert_walk_dict(PUBLISH_STANDARD, result)


def test_provenance_invoke():
    asset_list = []
    for index in range(0, 10):
        asset = DataAsset.create(f'test asset #{index+1}', secrets.token_hex(1024))
        asset_list.append(asset)

    activity_id = secrets.token_hex(32)
    agent_id = secrets.token_hex(32)

    inputs_text = 'test inputs text'
    outputs_text = 'test outputs text'
    result = create_invoke(activity_id, agent_id, asset_list, inputs_text, outputs_text)
    assert(isinstance(result, dict))
    fields = ['prefix', 'activity', 'entity', 'agent', 'wasAssociatedWith', 'wasGeneratedBy', 'wasDerivedFrom' ]
    for field in fields:
        assert(field in result)
    assert(len(fields) == len(result.keys()))


def test_assign_provenance_to_data_asset():
    agent_id = secrets.token_hex(32)
    test_data = secrets.token_hex(120)
    metadata = {
        METADATA_PROVENANCE_NAME: create_publish(agent_id)
    }
    asset = DataAsset.create('test data asset', test_data, metadata)
    assert(asset)


def test_assign_provenance_to_operation_asset():
    asset_list = []
    for index in range(0, 10):
        asset = DataAsset.create(f'test asset #{index+1}', secrets.token_hex(1024))
        asset_list.append(asset)

    activity_id = secrets.token_hex(32)
    agent_id = secrets.token_hex(32)
    inputs_text = json.dumps(
        {
            'asset_list': 'json'
        }
    )
    outputs_text =  json.dumps(
        {
            'asset_id': 'asset'
        }
    )
    metadata = OPERATION_METADATA
    metadata[METADATA_PROVENANCE_NAME] = create_invoke(activity_id, agent_id, asset_list, inputs_text, outputs_text)
    asset = OperationAsset(OPERATION_METADATA)
    assert(asset)
    assert(isinstance(asset, OperationAsset))



import pytest

TOKEN_AMOUNT_TO_TRANSFER = 2


def test_dispenser(network, accounts):
    dispenser_contract = network.get_contract('Dispenser')
    ocean_token_contract = network.get_contract('OceanToken')

    test_account = accounts[0]
    balance = ocean_token_contract.get_balance(test_account)
    tx_hash = dispenser_contract.request_tokens(test_account, TOKEN_AMOUNT_TO_TRANSFER)
    receipt = dispenser_contract.wait_for_receipt(tx_hash)

    new_balance = ocean_token_contract.get_balance(test_account)
    assert(balance + TOKEN_AMOUNT_TO_TRANSFER == new_balance)



"""

    Direct Purchase Contract

"""

from .contract_base import ContractBase

CONTRACT_NAME = 'DirectPurchase'
TOKEN_SENT_EVENT_NAME = 'TokenSent'


class DirectPurchaseContract(ContractBase):
    """Class representing the Token contract."""

    def __init__(self):
        ContractBase.__init__(self, CONTRACT_NAME)

    def send_token_and_log(self, account, to_account_address, amount, reference_1=None, reference_2=None):
        """
        Send tokens to address with tracking record in log.

        :param account: The Account to send the tokens from
        :type account: :class:`.Account` object
        :param address_to: Account address, str
        :param amount: token amount, int in unit of Ocean vodka's
        :param reference1: reference in log, int256
        :param reference2: reference in log, int256
        :return: void
        """

        if reference_1 is None:
            reference_1 = 0

        if reference_2 is None:
            reference_2 = 0

        to_address = self.get_account_address(to_account_address)
        amount_wei = self.to_wei(amount)

        tx_hash = self.call(
            'sendTokenAndLog',
            (
                to_address,
                amount_wei,
                self.web3.toBytes(hexstr=reference_1),
                self.web3.toBytes(hexstr=reference_2)
            ),
            account
        )
        return tx_hash

    def check_is_paid(self, from_account_address, to_account_address, amount, reference_1=None, reference_2=None):
        """
        Check if the log about transaction exists in blockchain.

        :param address_from: Account address sent the tokens from
        :param address_to: Account address, str
        :param amount: token amount, int in unit of Ocean vodka's
        :param reference: reference in log, int256
        :return: bool
        """

        from_address = self.get_account_address(from_account_address)
        to_address = self.get_account_address(to_account_address)
        amount_wei = self.to_wei(amount)

        argument_filters = {
            '_from': from_address,
            '_to': to_address,
            '_amount': amount_wei,
        }
        if reference_1:
            argument_filters['_reference1'] = self.web3.toBytes(hexstr=reference_1)

        if reference_2:
            argument_filters['_reference2'] = self.web3.toBytes(hexstr=reference_2)

        event_filter = self.create_event_filter(
            'TokenSent',
            None,
            from_block=1,
            argument_filters=argument_filters
        )

        if event_filter:
            for event_log in event_filter.get_all_entries():
                event_args = event_log.args
                if event_args._amount == amount_wei:
                    return True

        return False



"""
    test_04_agent_registration


    As a developer building or managing an Ocean Agent,
    I need to be able to register my Agent on the network and obtain an Agent ID

"""

from starfish.agent import RemoteAgent
from starfish.ddo import DDO



def test_04_agent_register_and_resolve(network, config, accounts):

    remote_agent = config.agent_list['remote']
    ddo = RemoteAgent.generate_ddo(remote_agent['url'])
    authentication_access = {
        'username': remote_agent['username'],
        'password': remote_agent['password']
    }

    register_account = accounts[0]

    did = ddo.did
    remote_agent = RemoteAgent.register(network, register_account, ddo, authentication_access)
    assert(remote_agent)
    found_ddo = DDO(json_text=network.resolve_did(did))
    assert(found_ddo.as_text() == ddo.as_text())


    resolved_agent = RemoteAgent.load(network, ddo.did, authentication_access=authentication_access)
    assert(resolved_agent)
    assert(resolved_agent.ddo)
    assert(resolved_agent.ddo.as_text() == ddo.as_text())



import pytest


from starfish.utils.did import did_generate_random
from starfish.agent import RemoteAgent
from starfish.agent.services import Services


def test_did_registry_contract(network, config, accounts):
    """

    Register and find a ddo based on a did

    """
    did_registry_contract = network.get_contract('DIDRegistry')

    did = did_generate_random()

    remote_agent = config.agent_list['remote']
    services = Services(remote_agent['url'], all_services=True)
    ddo = RemoteAgent.generate_ddo(services)
    ddo_text = ddo.as_text()

    register_account = accounts[0]

    tx_hash = did_registry_contract.register(register_account, did, ddo_text)
    receipt = did_registry_contract.wait_for_receipt(tx_hash)

    block_number = did_registry_contract.get_block_number(did)
    assert(receipt.blockNumber == block_number)

    saved_ddo_text = did_registry_contract.get_value(did)
    assert(ddo_text == saved_ddo_text)



"""

    Agent Manager, to load and resolve remote agents

"""
import logging

from starfish.agent.remote_agent import RemoteAgent
from starfish.ddo import DDO
from starfish.utils.did import (
    did_to_id,
    id_to_did
)

logger = logging.getLogger(__name__)


class AgentManager:
    def __init__(self, network):
        self._network = network
        self._items = {}

    def add(self, name, url=None, did=None, username=None, password=None):
        """

        Add a remote agent details to the list of remote agents managed by this class.

        :param str name: name of the remote agent
        :param str,dict url: url of the remote agent, or a dict if passing a dict of items
        :param str did: DID of the agent
        :param str username: Username access to the remote agent
        :param str password: Password access to the remote agent


        """
        if url and isinstance(url, dict):
            item = url
        else:
            item = {
                'url': url,
                'did': did,
                'username': username,
                'password': password
            }
        if item['url'] is None and item['did'] is None:
            raise ValueError('You must provide at least an URL or DID for the remote agent')
        self._items[name] = item

    def resolve_ddo(self, name):
        """

        Tries to resolve the remote agent to a ddo using it's provided URL or DID.

        :param str name: Name of the remote agent that was added to the managed list.
        :return dict: DDO text or None for the remote agent

        """

        if name not in self._items:
            raise ValueError(f'remote agent {name} not found in list')

        item = self._items[name]
        if item.get('ddo_text', None) is None:

            ddo_text = RemoteAgent.resolve_network_ddo(self._network, item.get('did', None))
            if ddo_text is None:
                authentication_access = None
                if item.get('username', None):
                    authentication_access = {
                        'username': item.get('username', None),
                        'password': item.get('password', None)
                    }
                ddo_text = RemoteAgent.resolve_url_ddo(item.get('url', None), authentication_access)

            self._items[name]['ddo_text'] = ddo_text
            self._items[name]['ddo'] = AgentManager.create_ddo_object(ddo_text)
        return self._items[name]['ddo_text']

    def load_agent(self, asset_agent_did_name):
        """

        Resolves and gets a valid remote agent for a given asset_did, agent_did or agent name

        :param str asset_agent_did_name:  Can be an asset_did, agent_did or name of the agent held by this object

        :return RemoteAgent: Object or None if none found

        """

        agent = None
        ddo_text = None

        # test for name in the agent list
        if asset_agent_did_name in self._items:
            logger.debug(f'found {asset_agent_did_name} in the list of agents')
            ddo_text = self.resolve_ddo(asset_agent_did_name)
        else:
            # find in list of did's
            # strip away the asset part of the id
            did_id = did_to_id(asset_agent_did_name)
            did = id_to_did(did_id)
            logger.debug(f'stripped possible did {asset_agent_did_name} to {did}')
            item = self.get_item_from_did(did)
            if item:
                logger.debug(f'found {did} in list of resolved items')
                ddo_text = item['ddo_text']
            else:
                # if it's a agent_did or asset_did
                ddo_text = RemoteAgent.resolve_network_ddo(self._network, did)
                if ddo_text:
                    logger.debug(f'resolved {did} from network')

        if ddo_text:
            ddo = DDO(json_text=ddo_text)
            authentication_access = None

            item = self.get_item_from_did(ddo.did)
            if item.get('username', None):
                authentication_access = {
                    'username': item.get('username', None),
                    'password': item.get('password', None)
                }
            agent = RemoteAgent(self._network, ddo_text, authentication_access)
        return agent

    def get_item_from_did(self, find_did):
        for name, item in self._items.items():
            did = None
            if 'ddo' in item:
                did = item['ddo'].did
            elif 'ddo_text' in item:
                ddo = AgentManager.create_ddo_object(item['ddo_text'])
                did = ddo.did
            if did and did == find_did:
                return item
        return None

    @property
    def items(self):
        return self._items

    @staticmethod
    def create_ddo_object(ddo_data):
        ddo = None
        if isinstance(ddo_data, str):
            ddo = DDO(json_text=ddo_data)
        elif isinstance(ddo_data, dict):
            ddo = DDO(dictionary=ddo_data)
        elif isinstance(ddo_data, DDO):
            ddo = ddo_data
        return ddo

    @staticmethod
    def get_did_from_ddo(ddo_data):
        ddo = AgentManager.create_ddo_object(ddo_data)
        if ddo:
            return ddo.did
        return None



from unittest.mock import Mock
import pytest
import secrets
import logging
import pathlib
import requests
from web3 import Web3, HTTPProvider

from tests.integration.libs.integration_test_config import IntegrationTestConfig

from starfish import DNetwork
from starfish.agent import RemoteAgent
from starfish.agent.services import Services
from starfish.account import Account

RESOURCES_PATH = pathlib.Path.cwd() / 'tests' / 'resources'
INTEGRATION_PATH = pathlib.Path.cwd() / 'tests' / 'integration'
CONFIG_FILE_PATH = RESOURCES_PATH / 'config_local.conf'

logging.getLogger('web3').setLevel(logging.INFO)
logging.getLogger('urllib3').setLevel(logging.INFO)

@pytest.fixture(scope='module')
def intergation_config():
    integration_test_config = IntegrationTestConfig(CONFIG_FILE_PATH)
    return integration_test_config

@pytest.fixture(scope="module")
def config(intergation_config):
    return intergation_config

@pytest.fixture(scope="module")
def network(config):
    network = DNetwork(config.network_url)
    return network


@pytest.fixture(scope="module")
def remote_agent(network, config):
    ddo_options = None
    remote_agent = config.agent_list['remote']
    services = Services(remote_agent['url'], all_services=True)
    ddo = RemoteAgent.generate_ddo(services)
    authentication_access = {
        'username': remote_agent['username'],
        'password': remote_agent['password'],
    }
    return RemoteAgent(network, ddo, authentication_access)

@pytest.fixture(scope='module')
def accounts(config):
    result = [
        Account(config.account_1),
        Account(config.account_2)
    ]
    return result


@pytest.fixture(scope='module')
def invokable_list(config):

    remote_agent = config.agent_list['remote']

    url = f'{remote_agent["url"]}/api/v1/admin/import-demo?id=operations'
    username = remote_agent['username']
    password = remote_agent['password']
    response = requests.post(url, auth=(username, password), headers={'accept':'application/json'})
    result = response.json()
    return result.get('invokables')



"""
    DID Lib to do DID's and DDO's
"""
import datetime
import json
import re
from base64 import (
    b64decode,
    b64encode
)

from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5
from web3 import Web3

from .authentication import Authentication
from .constants import (
    DID_DDO_CONTEXT_URL,
    KEY_PAIR_MODULUS_BIT
)
from .public_key_base import (
    PUBLIC_KEY_STORE_TYPE_PEM,
    PublicKeyBase
)
from .public_key_rsa import (
    AUTHENTICATION_TYPE_RSA,
    PUBLIC_KEY_TYPE_RSA,
    PublicKeyRSA
)
from .service import Service


class DDO:
    """
    DDO class to create, import, export, validate DDO objects.

    """

    def __init__(self, did='', json_text=None, json_filename=None, created=None, dictionary=None):
        """ clear the DDO data values """
        self._did = ''
        self._public_keys = []
        self._authentications = []
        self._services = []
        self._proof = None
        self._created = None

        self._did = did
        if created:
            self._created = created
        else:
            self._created = DDO.get_timestamp()

        if json_filename:
            with open(json_filename, 'r') as file_handle:
                json_text = file_handle.read()

        if json_text:
            self._read_dict(json.loads(json_text))

        if dictionary:
            self._read_dict(dictionary)

    def add_public_key(self, public_key):
        """add a public key object to the list of public keys"""
        self._public_keys.append(public_key)

    def add_authentication(self, key_id, authentication_type=None):
        """add a authentication public key id and type to the list of authentications"""
        if isinstance(key_id, Authentication):
            # adding an authentication object
            authentication = key_id
        elif isinstance(key_id, PublicKeyBase):
            public_key = key_id
            # this is going to be a embedded public key
            authentication = Authentication(public_key, public_key.authentication_type)
        else:
            # with key_id as a string, we also need to provide the authentication type
            if authentication_type is None:
                raise ValueError
            authentication = Authentication(key_id, authentication_type)

        self._authentications.append(authentication)

    def add_signature(self, public_key_store_type=PUBLIC_KEY_STORE_TYPE_PEM, is_embedded=False):
        """add a signature with a public key and authentication entry for validating this DDO
        returns the private key as part of the private/public key pair"""

        key_pair = RSA.generate(KEY_PAIR_MODULUS_BIT, e=65537)
        public_key_raw = key_pair.publickey()
        private_key_pem = key_pair.exportKey("PEM")

        # find the current public key count
        next_index = self.public_key_count + 1
        key_id = '{0}#keys={1}'.format(self._did, next_index)

        public_key = PublicKeyRSA(key_id, owner=key_id)

        public_key.set_encode_key_value(public_key_raw, public_key_store_type)

        if is_embedded:
            # also add the authentication key as embedded key with the authentication
            self.add_authentication(public_key)
        else:
            # add the public key to the DDO list of public keys
            self.add_public_key(public_key)

            # add the public key id and type for this key to the authentication
            self.add_authentication(public_key.id, public_key.authentication_type)

        return private_key_pem

    def add_service(self, service_type, service_endpoint=None, service_id=None, values=None):
        """add a service to the list of services on the DDO"""
        if isinstance(service_type, Service):
            service = service_type
        else:
            if service_id is None:
                service_id = self._did
            service = Service(service_id, service_endpoint, service_type, values)
        self._services.append(service)

    def as_text(self, is_proof=True, is_pretty=False):
        """return the DDO as a JSON text
        if is_proof == False then do not include the 'proof' element"""
        data = self.as_dictionary(is_proof)
        if is_pretty:
            return json.dumps(data, indent=2, separators=(',', ': '))

        return json.dumps(data)

    def as_dictionary(self, is_proof=True):
        if self._created is None:
            self._created = DDO.get_timestamp()

        data = {
            '@context': DID_DDO_CONTEXT_URL,
            'id': self._did,
            'created': self._created,
        }
        if self._public_keys:
            values = []
            for public_key in self._public_keys:
                values.append(public_key.as_dictionary)
            data['publicKey'] = values
        if self._authentications:
            values = []
            for authentication in self._authentications:
                values.append(authentication.as_dictionary)
            data['authentication'] = values
        if self._services:
            values = []
            for service in self._services:
                values.append(service.as_dictionary)
            data['service'] = values
        if self._proof and is_proof:
            data['proof'] = self._proof

        return data

    def _read_dict(self, dictionary):
        """import a JSON dict into this DDO"""
        values = dictionary
        self._did = values['id']
        self._created = values.get('created', None)
        if 'publicKey' in values:
            self._public_keys = []
            for value in values['publicKey']:
                if isinstance(value, str):
                    value = json.loads(value)
                self._public_keys.append(DDO.create_public_key_from_json(value))
        if 'authentication' in values:
            self._authentications = []
            for value in values['authentication']:
                if isinstance(value, str):
                    value = json.loads(value)
                self._authentications.append(DDO.create_authentication_from_json(value))
        if 'service' in values:
            self._services = []
            for value in values['service']:
                if isinstance(value, str):
                    value = json.loads(value)
                self.services.append(DDO.create_service_from_json(value, self._did))
        if 'proof' in values:
            self._proof = values['proof']

    def add_proof(self, authorisation_index, private_key=None, signature_text=None):
        """add a proof to the DDO, based on the public_key id/index and signed with the private key
        add a static proof to the DDO, based on one of the public keys"""

        # find the key using an index, or key name
        if isinstance(authorisation_index, dict):
            self._proof = authorisation_index
            return

        if private_key is None:
            raise ValueError

        authentication = self._authentications[authorisation_index]
        if not authentication:
            raise IndexError
        if authentication.is_public_key:
            sign_key = authentication.public_key
        else:
            sign_key = self.get_public_key(authentication.public_key_id)

        if sign_key is None:
            raise IndexError

        # get the signature text if not provided

        if signature_text is None:
            hash_text_list = self.hash_text_list
            signature_text = "".join(hash_text_list)

        # just incase clear out the current static proof property
        self._proof = None

        signature = DDO.sign_text(signature_text, private_key, sign_key.type)

        self._proof = {
            'type': sign_key.type,
            'created': DDO.get_timestamp(),
            'creator': sign_key.id,
            'signatureValue': b64encode(signature).decode('utf-8'),
        }

    def validate_proof(self, signature_text=None):
        """validate the static proof created with this DDO, return True if valid
        if no static proof exists then return False"""

        if not signature_text:
            hash_text_list = self.hash_text_list()
            signature_text = "".join(hash_text_list)
        if self._proof is None:
            return False
        if not isinstance(self._proof, dict):
            return False

        if 'creator' in self._proof and 'signatureValue' in self._proof:
            signature = b64decode(self._proof['signatureValue'])
            return self.validate_from_key(self._proof['creator'], signature_text, signature)
        return False

    @property
    def is_proof_defined(self):
        """return true if a static proof exists in this DDO"""
        return self._proof is not None

    def validate_from_key(self, key_id, signature_text, signature_value):
        """validate a signature based on a given public_key key_id/name """

        public_key = self.get_public_key(key_id, True)
        if public_key is None:
            return False

        key_value = public_key.decode_value
        if key_value is None:
            return False

        authentication = self.get_authentication_from_public_key_id(public_key.id)
        if authentication is None:
            return False

        # if public_key.get_store_type() != PUBLIC_KEY_STORE_TYPE_PEM:
        # key_value = key_value.decode()

        return DDO.validate_signature(signature_text, key_value, signature_value, authentication.type)

    def get_public_key(self, key_id, is_search_embedded=False):
        """key_id can be a string, or int. If int then the index in the list of keys"""
        if isinstance(key_id, int):
            return self._public_keys[key_id]

        for item in self._public_keys:
            if item.id == key_id:
                return item

        if is_search_embedded:
            for authentication in self._authentications:
                if authentication.public_key_id == key_id:
                    return authentication.public_key
        return None

    @property
    def public_key_count(self):
        """return the count of public keys in the list and embedded"""
        index = len(self._public_keys)
        for authentication in self._authentications:
            if authentication.is_public_key:
                index += 1
        return index

    def get_authentication_from_public_key_id(self, key_id):
        """return the authentication based on it's id"""
        for authentication in self._authentications:
            if authentication.is_key_id(key_id):
                return authentication
        return None

    def get_service(self, service_type=None, service_id=None):
        """return a service using"""
        for service in self._services:
            if service.id == service_id and service_id:
                return service
            if service.type == service_type and service_type:
                return service
        return None

    def find_service_by_key_value(self, service_key, value):
        for s in self._services:
            if service_key in s.values and s.values[service_key] == value:
                return s

        return None

    def validate(self):
        """validate the ddo data structure"""
        if self._public_keys and self._authentications:
            for authentication in self._authentications:
                if not authentication.is_valid:
                    return False
                if authentication.is_public_key:
                    public_key = authentication.public_key
                else:
                    key_id = authentication.public_key_id()
                    public_key = self.get_public_key(key_id)
                if not public_key.is_valid:
                    return False
        if self._services:
            for service in self._services:
                if not service.is_valid:
                    return False

        # validate if proof defined in this DDO
        if self.is_proof_defined:
            if not self.validate_proof:
                return False
        return True

    @property
    def hash_text_list(self):
        """return a list of all of the hash text"""
        hash_text = []
        if self._created:
            hash_text.append(self._created)

        if self._public_keys:
            for public_key in self._public_keys:
                if public_key.type:
                    hash_text.append(public_key.type)
                if public_key.value:
                    hash_text.append(public_key.value)

        if self._authentications:
            for authentication in self._authentications:
                if authentication.is_public_key:
                    public_key = authentication.public_key
                    if public_key.type:
                        hash_text.append(public_key.type)
                    if public_key.value:
                        hash_text.append(public_key.value)

        if self._services:
            for service in self._services:
                hash_text.append(service.type)
                hash_text.append(service.endpoint)

        # if no data can be found to hash then raise an error
        if not hash_text:
            raise ValueError
        return hash_text

    def calculate_hash(self):
        """return a sha3 hash of important bits of the DDO, excluding any DID portion,
        as this hash can be used to generate the DID"""
        hash_text_list = self.hash_text_list()
        return Web3.sha3(text="".join(hash_text_list))

    @property
    def is_empty(self):
        """return True if this DDO object is empty"""
        return self._did == '' \
            and len(self._public_keys) == 0 \
            and len(self._authentications) == 0 \
            and len(self._services) == 0 \
            and self._proof is None \
            and self._created is None

    @property
    def is_did_assigend(self):
        """return true if a DID is assigned to this DDO"""
        return self._did != '' and self._did is not None

    @property
    def created_time(self):
        """return the DDO created time, can be None"""
        return self._created

    def create_new(self, did):
        """method to copy a DDO and assign a new did to all of the keys to an empty/non DID assigned DDO.
        we assume that this ddo has been created as empty ( no did )"""

        if self.is_did_assigend:
            raise Exception('Cannot assign a DID to a completed DDO object')
        ddo = DDO(did, created=self._created)
        for public_key in self._public_keys:
            public_key.assign_did(did)
            ddo.add_public_key(public_key)

        for authentication in self._authentications:
            authentication.assign_did(did)
            ddo.add_authentication(authentication)

        for service in self._services:
            service.assign_did(did)
            ddo.add_service(service)

        if self.is_proof_defined:
            if re.match('^#.*', self._proof['creator']):
                proof = self._proof
                proof['creator'] = did + proof['creator']
            ddo.add_proof(proof)

        return ddo

    @property
    def did(self):
        """ get the DID """
        return self._did

    @property
    def public_keys(self):
        """get the list of public keys"""
        return self._public_keys

    @property
    def authentications(self):
        """get the list authentication records"""
        return self._authentications

    @property
    def services(self):
        """get the list of services"""
        return self._services

    @property
    def proof(self):
        """ get the static proof, or None """
        return self._proof

    @property
    def is_valid(self):
        """return True if this DDO is valid"""
        return self.validate()

    @staticmethod
    def sign_text(text, private_key, sign_type=PUBLIC_KEY_TYPE_RSA):
        """Sign some text using the private key provided"""
        signed_text = None
        if sign_type == PUBLIC_KEY_TYPE_RSA:
            signer = PKCS1_v1_5.new(RSA.import_key(private_key))
            text_hash = SHA256.new(text.encode('utf-8'))
            signed_text = signer.sign(text_hash)
        else:
            raise NotImplementedError
        return signed_text

    @staticmethod
    def validate_signature(text, key, signature, sign_type=AUTHENTICATION_TYPE_RSA):
        """validate a signature based on some text, public key and signature"""
        result = False
        try:
            if sign_type == AUTHENTICATION_TYPE_RSA:
                rsa_key = RSA.import_key(key)
                verifier = PKCS1_v1_5.new(rsa_key)
                if verifier:
                    text_hash = SHA256.new(text.encode('utf-8'))
                    result = verifier.verify(text_hash, signature)
            else:
                raise NotImplementedError
        except (ValueError, TypeError):
            result = False

        return result

    @staticmethod
    def create_public_key_from_json(values):
        """create a public key object based on the values from the JSON record"""
        # currently we only support RSA public keys
        public_key = PublicKeyRSA(values['id'], owner=values.get('owner', None))
        public_key.set_key_value(values)
        return public_key

    @staticmethod
    def create_authentication_from_json(values):
        """create authentitaciton object from a JSON string"""
        key_id = values['publicKey']
        authentication_type = values['type']
        if isinstance(key_id, dict):
            public_key = DDO.create_public_key_from_json(key_id)
            authentication = Authentication(public_key, public_key.authentication_type)
        else:
            authentication = Authentication(key_id, authentication_type)

        return authentication

    @staticmethod
    def create_service_from_json(values, service_id):
        """create a service object from a JSON string"""

        if 'id' in values:
            service_id = values['id']
        if 'serviceEndpoint' not in values:
            raise IndexError
        if 'type' not in values:
            raise IndexError
        service = Service(service_id, values['serviceEndpoint'], values['type'], values)
        return service

    @staticmethod
    def get_timestamp():
        """return the current system timestamp"""
        return str(datetime.datetime.now())



"""

    Test DNetwork class

"""
import pytest
import secrets

from starfish.utils.did import did_generate_random
from starfish.agent import RemoteAgent
from starfish.agent.services import Services

from starfish import DNetwork

TEST_AMOUNT = 5

def test_dnetwork_basic(network, config):
    assert(network.name == 'spree')
    assert(network.web3)
    assert(network.contract_names)
    assert(network.url == config.network_url)

def test_dnetwork_account(network, accounts):
    test_account = accounts[0]
    ether_balance = network.get_ether_balance(test_account)
    assert(ether_balance)

    token_balance = network.get_token_balance(test_account)
    assert(token_balance)

def test_dnetwork_request_test_tokens(network, accounts):

    test_account = accounts[0]

    token_balance = network.get_token_balance(test_account)
    assert(token_balance)

    network.request_test_tokens(test_account, TEST_AMOUNT)

    new_token_balance = network.get_token_balance(test_account)
    assert(new_token_balance)
    assert(token_balance + TEST_AMOUNT == new_token_balance)

def test_dnetwork_send_ether(network, accounts):
    from_account = accounts[0]
    to_account = accounts[1]

    from_balance = network.get_ether_balance(from_account)
    to_balance = network.get_ether_balance(to_account)

    receipt = network.send_ether(from_account, to_account, TEST_AMOUNT)

    new_from_balance = network.get_ether_balance(from_account)
    new_to_balance = network.get_ether_balance(to_account)

    assert(from_balance - TEST_AMOUNT == new_from_balance)
    assert(to_balance + TEST_AMOUNT == new_to_balance)


def test_dnetwork_send_token(network, accounts):
    from_account = accounts[0]
    to_account = accounts[1]

    network.request_test_tokens(from_account, TEST_AMOUNT)

    from_balance = network.get_token_balance(from_account)
    to_balance = network.get_token_balance(to_account)

    receipt = network.send_token(from_account, to_account, TEST_AMOUNT)

    new_from_balance = network.get_token_balance(from_account)
    new_to_balance = network.get_token_balance(to_account)

    assert(from_balance - TEST_AMOUNT == new_from_balance)
    assert(to_balance + TEST_AMOUNT == new_to_balance)

def test_dnetwork_send_token_and_log(network, accounts):
    from_account = accounts[0]
    to_account = accounts[1]

    network.request_test_tokens(from_account, TEST_AMOUNT)

    from_balance = network.get_token_balance(from_account)
    to_balance = network.get_token_balance(to_account)

    ref_1 = secrets.token_hex(32)
    ref_2 = secrets.token_hex(32)
    receipt = network.send_token_and_log(from_account, to_account, TEST_AMOUNT, ref_1, ref_2)

    new_from_balance = network.get_token_balance(from_account)
    new_to_balance = network.get_token_balance(to_account)

    assert(from_balance - TEST_AMOUNT == new_from_balance)
    assert(to_balance + TEST_AMOUNT == new_to_balance)

    is_sent = network.is_token_sent(
        from_account,
        to_account,
        TEST_AMOUNT,
        ref_1,
        ref_2
    )
    assert(is_sent)

def test_dnetwork_regiser_provenance(network, accounts):

    register_account = accounts[0]
    asset_id = secrets.token_hex(32)

    network.register_provenace(register_account, asset_id)
    event_list = network.get_provenace_event_list(asset_id)
    assert(len(event_list) > 0)


def test_dnetwork_regiser_resolve_did_ddo(config, network, accounts):
    did = did_generate_random()

    remote_agent = config.agent_list['remote']
    services = Services(remote_agent['url'], all_services=True)
    ddo = RemoteAgent.generate_ddo(services)
    ddo_text = ddo.as_text()

    register_account = accounts[0]
    network.register_did(register_account, did, ddo_text)

    ddo_text_saved = network.resolve_did(did)

    assert(ddo_text == ddo_text_saved)








"""
    test_07_metadata_access

    As a developer working with Ocean,
    I need a way to request metadata for an arbitrary asset

"""

import secrets
import json

from starfish.asset import DataAsset
from starfish.agent import RemoteAgent


def test_07_metadata_access(resources, remote_agent):
    test_data = secrets.token_bytes(1024)
    asset_data = DataAsset.create('TestAsset', test_data)
    asset = remote_agent.register_asset(asset_data)
    assert(asset.asset_id)
    assert(asset.metadata)
    assert(asset.metadata == asset_data.metadata)




import pytest
import secrets
from web3 import Web3

from starfish.account import Account


def test_init(config):
    account = Account(config.accounts[0].as_dict)
    assert(account)
    assert(account.address)

    account = Account(config.accounts[0].as_tuple)
    assert(account)
    assert(account.address)

    account = Account(config.accounts[0].as_list)
    assert(account)
    assert(account.address)

    account = Account(config.accounts[0].address, 'test-password', 'test-keyfie')
    assert(account)
    assert(account.address)

def test_is_address_equal(config):
    account = Account(config.accounts[0].as_dict)
    assert(account.is_address_equal(config.accounts[0].test_address))

def test_is_valid(config):
    account = Account(config.accounts[0].as_dict)
    assert(account.is_valid)

def test_is_password(config):
    account = Account(config.accounts[0].as_dict)
    assert(account.password == config.accounts[0].test_password)

def test_is_keyfile(config):
    account = Account(config.accounts[0].as_dict)
    assert(account.keyfile == config.accounts[0].test_keyfile)

def test_address(config):
    account = Account(config.accounts[0].as_dict)
    assert(account.address.lower() == config.accounts[0].test_address.lower())

def test_as_checksum_address(config):
    account = Account(config.accounts[0].as_dict)
    assert(account.as_checksum_address == Web3.toChecksumAddress(config.accounts[0].test_address))

def test_password(config):
    account = Account(config.accounts[0].as_dict)
    assert(account.password == config.accounts[0].test_password)

def test_set_password(config):
    account = Account(config.accounts[0].as_dict)
    new_password = secrets.token_hex(48)
    account.set_password(new_password)
    assert(account.password != config.accounts[0].test_password)
    assert(account.password == new_password)




"""
    Public key HEX

    This is originally from
    https://github.com/oceanprotocol/squid-py/tree/b62b1bae303c499ca43f09dad1677001d9be840e/squid_py/ddo
"""

from .public_key_base import (
    PUBLIC_KEY_STORE_TYPE_HEX,
    PublicKeyBase
)

AUTHENTICATION_TYPE_HEX = 'HexVerificationKey'
PUBLIC_KEY_TYPE_HEX = 'PublicKeyHex'


class PublicKeyHex(PublicKeyBase):
    """Encode key value using Hex."""

    def __init__(self, key_id, **kwargs):
        PublicKeyBase.__init__(self, key_id, **kwargs)
        self._type = PUBLIC_KEY_TYPE_HEX
        self._store_type = PUBLIC_KEY_STORE_TYPE_HEX

    @property
    def authentication_type(self):
        """Return the type of authentication supported by this class."""
        return AUTHENTICATION_TYPE_HEX
